from typing import Any, Literal, Union, overload

import numpy as np
from numpy.typing import ArrayLike, DTypeLike, NDArray

@overload
def unique(
    labels: ArrayLike,
    return_index: Literal[False] = False,
    return_inverse: Literal[False] = False,
    return_counts: Literal[False] = False,
    axis: Union[int, None] = None,
) -> NDArray[Any]: ...
@overload
def unique(
    labels: ArrayLike,
    return_index: Literal[True],
    return_inverse: Literal[False] = False,
    return_counts: Literal[False] = False,
    axis: Union[int, None] = None,
) -> tuple[NDArray[Any], NDArray[Any]]: ...
@overload
def unique(
    labels: ArrayLike,
    return_index: Literal[False],
    return_inverse: Literal[True],
    return_counts: Literal[False] = False,
    axis: Union[int, None] = None,
) -> tuple[NDArray[Any], NDArray[Any]]: ...
@overload
def unique(
    labels: ArrayLike,
    return_index: Literal[False] = False,
    *,
    return_inverse: Literal[True],
    return_counts: Literal[False] = False,
    axis: Union[int, None] = None,
) -> tuple[NDArray[Any], NDArray[Any]]: ...
@overload
def unique(
    labels: ArrayLike,
    return_index: Literal[False],
    return_inverse: Literal[False],
    return_counts: Literal[True],
    axis: Union[int, None] = None,
) -> tuple[NDArray[Any], NDArray[Any]]: ...
@overload
def unique(
    labels: ArrayLike,
    return_index: Literal[False] = False,
    return_inverse: Literal[False] = False,
    *,
    return_counts: Literal[True],
    axis: Union[int, None] = None,
) -> tuple[NDArray[Any], NDArray[Any]]: ...
@overload
def unique(
    labels: ArrayLike,
    return_index: Literal[True],
    return_inverse: Literal[True],
    return_counts: Literal[False] = False,
    axis: Union[int, None] = None,
) -> tuple[NDArray[Any], NDArray[Any], NDArray[Any]]: ...
@overload
def unique(
    labels: ArrayLike,
    return_index: Literal[True],
    return_inverse: Literal[False],
    return_counts: Literal[True],
    axis: Union[int, None] = None,
) -> tuple[NDArray[Any], NDArray[Any], NDArray[Any]]: ...
@overload
def unique(
    labels: ArrayLike,
    return_index: Literal[True],
    return_inverse: Literal[False] = False,
    *,
    return_counts: Literal[True],
    axis: Union[int, None] = None,
) -> tuple[NDArray[Any], NDArray[Any], NDArray[Any]]: ...
@overload
def unique(
    labels: ArrayLike,
    return_index: Literal[False],
    return_inverse: Literal[True],
    return_counts: Literal[True],
    axis: Union[int, None] = None,
) -> tuple[NDArray[Any], NDArray[Any], NDArray[Any]]: ...
@overload
def unique(
    labels: ArrayLike,
    return_index: Literal[False] = False,
    *,
    return_inverse: Literal[True],
    return_counts: Literal[True],
    axis: Union[int, None] = None,
) -> tuple[NDArray[Any], NDArray[Any], NDArray[Any]]: ...
@overload
def unique(
    labels: ArrayLike,
    return_index: Literal[True],
    return_inverse: Literal[True],
    return_counts: Literal[True],
    axis: Union[int, None] = None,
) -> tuple[NDArray[Any], NDArray[Any], NDArray[Any], NDArray[Any]]: ...
def unique(  # type: ignore[misc]
    labels: ArrayLike,
    return_index: bool = False,
    return_inverse: bool = False,
    return_counts: bool = False,
    axis: Union[int, None] = None,
) -> Union[
    NDArray[Any],
    tuple[NDArray[Any], NDArray[Any]],
    tuple[NDArray[Any], NDArray[Any], NDArray[Any]],
    tuple[NDArray[Any], NDArray[Any], NDArray[Any], NDArray[Any]],
]:
    """Compute the sorted set of unique labels in the input array.

    Args:
        labels: The input array containing labels.
        return_index: If True, also return the index of the first detected
            occurance of each label.
        return_inverse: If True, also return the indices of the unique array
            (for the specified axis, if provided) that can be used to
            reconstruct the input array.
        return_counts: If True, also return the unique label frequency as an
            array.
        axis: If specified and not None, compute the unique values along this
            axis.

    Returns:
        Either an array of the sorted values or a tuple containing the following
        elements depending on the flags set.

        unique:
            The sorted unique values.
            Always provided.

        unique_indices, optional:
            The indices of the first occurrences of the unique values in the
            original array.
            Only provided if return_index is True.

        unique_inverse, optional:
            The indices to reconstruct the original array from the unique array.
            Only provided if return_inverse is True.

        unique_counts, optional:
            The number of times each of the unique values comes up in the
            original array.
            Only provided if return_counts is True.
    """
    ...

def renumber(
    arr: NDArray[Any],
    start: Union[int, float] = 1,
    preserve_zero: bool = True,
    in_place: bool = False,
) -> Union[tuple[NDArray[Any], dict[int, int], dict[float, float]]]:
    """Renumber an array.

    Given an array of integers, renumber all the unique values starting
    from 1. This can allow us to reduce the size of the data width required
    to represent it.

    Args:
        arr: A numpy array.
        start (default: 1): Start renumbering from this value.
        preserve_zero (default: True): Don't renumber zero.
        in_place (default: False): Perform the renumbering in-place to avoid
            an extra copy. This option depends on a fortran or C contiguous
            array. A copy will be made if the array is not contiguous.

    Returns:
        A renumbered array, dict with remapping of oldval => newval.
    """
    ...

def indices(
    arr: NDArray[Any],
    value: Union[int, float],
) -> NDArray[Any]:
    """Search an array for indices where value matches the array value."""
    ...

def remap(
    arr: ArrayLike,
    table: Union[dict[int, int], dict[float, float]],
    preserve_missing_labels: bool = False,
    in_place: bool = False,
) -> NDArray[Any]:
    """Remap an input numpy array in-place according to a dictionary "table".

    Args:
        arr: An N-dimensional numpy array.
        table: A dictionary resembling: { label: new_label_value, ... }.
        preserve_missing_labels: If an array value is not present in "table"...
            True: Leave it alone.
            False: Throw a KeyError.
        in_place: if True, modify the input array to reduce memory consumption.

    Returns:
        The remapped array.
    """
    ...

def refit(
    arr: NDArray[Any],
    value: Union[int, float, None] = None,
    increase_only: bool = False,
    exotics: bool = False,
) -> NDArray[Any]:
    """Resize the array to the smallest dtype of the same kind that will fit.

    For example, if the input array is uint8 and the value is 2^20 return the
    array as a uint32.

    Works for standard floating, integer, unsigned integer, and complex types.

    Args:
        arr: A numpy array.
        value: Value to fit array to. If None, it is set to the value of the
            absolutely larger of the min and max value in the array.
        increase_only: if true, only resize the array if it can't contain value.
            If false, always resize to the smallest size that fits.
        exotics: If true, allow e.g. half precision floats (16-bit) or double
            complex (128-bit)

    Returns:
        The refitted array.
    """
    ...

def narrow_dtype(
    dtype: DTypeLike,
    exotics: bool = False,
) -> DTypeLike:
    """Widen the given dtype to the next size of the same type.

    For example, int16 -> int8 or uint64 -> uint32.

    8-bit types will map to themselves.

    Args:
        exotics: Whether to include exotics like half precision floats (16-bit)
            or double complex (128-bit).

    Returns:
        The downgraded dtype.
    """
    ...

def widen_dtype(
    dtype: DTypeLike,
    exotics: bool = False,
) -> DTypeLike:
    """Widen the given dtype to the next size of the same type.

    For example, int8 -> int16 or uint32 -> uint64.

    64-bit types will map to themselves.

    Args:
        exotics: Whether to include exotics like half precision floats (16-bit)
            or double complex (128-bit).

    Returns:
        The upgraded dtype.
    """
    ...

def mask(
    arr: ArrayLike,
    labels: ArrayLike,
    in_place: bool = False,
    value: Union[int, float] = 0,
) -> NDArray[Any]:
    """Mask out designated labels in an array with the given value.

    Alternative implementation of:

        arr[np.isin(labels)] = value

    Args:
        arr: An N-dimensional numpy array.
        labels: An iterable list of integers.
        in_place: If True, modify the input array to reduce memory consumption.
        value: A mask value.

    Returns:
        The array with `labels` masked out.
    """
    ...

def mask_except(
    arr: NDArray[Any],
    labels: ArrayLike,
    in_place: bool = False,
    value: Union[int, float] = 0,
) -> NDArray[Any]:
    """Mask out all labels except the provided list.

    Alternative implementation of:

        arr[~np.isin(labels)] = value

    Args:
        arr: An N-dimensional numpy array.
        labels: An iterable list of integers.
        in_place: If True, modify the input array to reduce memory consumption.
        value: A mask value.

    Returns:
        The array with all labels except `labels` masked out.
    """
    ...

def component_map(
    component_labels: ArrayLike,
    parent_labels: ArrayLike,
) -> Union[dict[int, int], dict[float, float]]:
    """Generate a mapping from connected components to their parent labels.

    Given two sets of images that have a surjective mapping between their
    labels, generate a dictionary for that mapping. For example, generate a
    mapping from connected components of labels to their parent labels.

    Returns:
        { $COMPONENT_LABEL: $PARENT_LABEL }

    Examples:
        >>> fastremap.component_map([1, 2, 3, 4], [5, 5, 6, 7])
        {1: 5, 2: 5, 3: 6, 4: 7}
    """
    ...

def inverse_component_map(
    parent_labels: ArrayLike,
    component_labels: ArrayLike,
) -> Union[dict[int, int], dict[float, float]]:
    """Generate a mapping from parent labels to connected components.

    Given two sets of images that have a mapping between their labels, generate
    a dictionary for that mapping. For example, generate a mapping from
    connected components of labels to their parent labels.

    Returns:
        A dictionary resembling: { $PARENT_LABEL: [ $COMPONENT_LABELS, ... ] }.

    Examples:
        >>> fastremap.inverse_component_map([1, 2, 1, 3], [4, 4, 5, 6])
        {1: [4, 5], 2: [4], 3: [6]}
    """
    ...

def remap_from_array(
    arr: NDArray[np.uint],
    vals: NDArray[np.uint],
    in_place: bool = True,
) -> NDArray[Any]:
    """Remap an input numpy array according to the given values array.

    Args:
        arr: An N-dimensional numpy array.
        vals: An array of values to remap to, where the index of the value in
            the array corresponds to the label in the input array.
        in_place: If True, modify the input array to reduce memory consumption.

    Returns:
        The remapped array.
    """
    ...

def remap_from_array_kv(
    arr: NDArray[np.integer],
    keys: NDArray[np.integer],
    vals: NDArray[np.integer],
    preserve_missing_labels: bool = True,
    in_place: bool = True,
) -> NDArray[Any]:
    """Remap an input numpy array according to the keys and values arrays.

    Args:
        arr: An N-dimensional numpy array.
        keys: An array of keys to remap from. Must be the same length as `vals`.
        vals: An array of values to remap to. Must be the same length as `vals`.
        preserve_missing_labels: If an array value is not present in `keys`...
            True: Leave it alone.
            False: Throw a KeyError.
        in_place: If True, modify the input array to reduce memory consumption.

    Returns:
        The remapped array.
    """
    ...

def transpose(arr: NDArray[Any]) -> NDArray[Any]:
    """For up to four dimensional matrices, perform in-place transposition.

    Square matrices up to three dimensions are faster than numpy's out-of-place
    algorithm. Default to the out-of-place implementation numpy uses for cases
    that aren't specially handled.

    Args:
        arr: The input numpy array to transpose.

    Returns:
        The transposed numpy array
    """
    ...

def asfortranarray(arr: NDArray[Any]) -> NDArray[Any]:
    """For up to four dimensional matrices, perform in-place transposition.

    Square matrices up to three dimensions are faster than numpy's out-of-place
    algorithm. Default to the out-of-place implementation numpy uses for cases
    that aren't specially handled.

    Args:
        arr: The input numpy array to transpose.

    Returns:
        The transposed numpy array.
    """
    ...

def ascontiguousarray(arr: NDArray[Any]) -> NDArray[Any]:
    """For up to four dimensional matrices, perform in-place transposition.

    Square matrices up to three dimensions are faster than numpy's out-of-place
    algorithm. Default to the out-of-place implementation numpy uses for cases
    that aren't specially handled.

    Args:
        arr: The input numpy array to transpose.

    Returns:
        The transposed numpy array.
    """
    ...

def minmax(
    arr: NDArray[Any],
) -> tuple[Union[int, float, None], Union[int, float, None]]:
    """Returns (min(arr), max(arr)) computed in a single pass.

    Returns (None, None) if array is size zero.
    """
    ...

def pixel_pairs(labels: NDArray[Any]) -> int:
    """Computes the number of matching adjacent memory locations.

    This is useful for rapidly evaluating whether an image is
    more binary or more connectomics like.
    """
    ...

def foreground(arr: NDArray[Any]) -> int:
    """Returns the number of non-zero voxels in an array."""
    ...

def point_cloud(arr: NDArray[Any]) -> dict[int, NDArray[Any]]:
    """Generate a mapping from labels to their (x, y, z) position in the image.

    Zero is considered a background label.

    Args:
        arr: A 2D or 3D numpy array.

    Returns:
        A dictionary mapping label values to their (x, y, z) coordinates in the
        image. The coordinates are stored as a numpy array of shape (N, 3),
        where N is the number of points for that label.
    """
    ...

def tobytes(
    image: NDArray[Any],
    chunk_size: tuple[int, int, int],
    order: str = "C",
) -> list[bytes]:
    """Compute the bytes with the image divided into a grid of cutouts.

    Return the resultant binaries indexed by their cutout's gridpoint in
    fortran order.

    This is faster than calling tobytes on each cutout individually if the input
    and output orders match.
    """
    ...



import pytest

import numpy as np

import fastremap

DTYPES = (
  np.uint8, np.uint16, np.uint32, np.uint64,
  np.int8, np.int16, np.int32, np.int64
)

def test_empty_renumber():
  for dtype in DTYPES:
    data = np.array([], dtype=dtype)
    data2, remapdict = fastremap.renumber(data, preserve_zero=False)

    assert np.all(data2 == [])
    assert remapdict == {}

def test_1d_renumber():
  for dtype in DTYPES:
    print(dtype)
    data = np.arange(8).astype(dtype)
    data = np.flip(data)

    data2 = np.copy(data)
    data2, remapdict = fastremap.renumber(data2, preserve_zero=False)

    assert np.all(data2 == np.arange(1,9))
    assert len(remapdict) > 0

    data2 = np.copy(data)
    data2, remapdict = fastremap.renumber(data2, preserve_zero=True)

    assert data2[-1] == 0
    assert np.all(data2 == [1,2,3,4,5,6,7,0])
    assert len(remapdict) > 0

  data = np.arange(8).astype(bool)
  data = np.flip(data)

  data2 = np.copy(data)
  data2, remapdict = fastremap.renumber(data2, preserve_zero=False)

  assert np.all(data2 == [1,1,1,1,1,1,1,2])
  assert len(remapdict) > 0

  data2 = np.copy(data)
  data2, remapdict = fastremap.renumber(data2, preserve_zero=True)

  assert np.all(data2 == [1,1,1,1,1,1,1,0])
  assert len(remapdict) > 0

def test_2d_renumber():
  for dtype in DTYPES:
    data = np.array([
      [ 5,  5,  5, 2],
      [ 3,  5,  5, 0],
      [ 1,  2,  4, 1],
      [20, 19, 20, 1],
    ], dtype=dtype)

    data2 = np.copy(data, order='C')
    data2, remapdict = fastremap.renumber(data2, preserve_zero=True)

    assert np.all(data2 == [
      [1, 1, 1, 2],
      [3, 1, 1, 0],
      [4, 2, 5, 4],
      [6, 7, 6, 4],
    ])

    data2 = np.copy(data, order='F')
    data2, remapdict = fastremap.renumber(data2, preserve_zero=True)

    assert np.all(data2 == [
      [1, 1, 1, 5],
      [2, 1, 1, 0],
      [3, 5, 7, 3],
      [4, 6, 4, 3],
    ])

@pytest.mark.parametrize("dtype", DTYPES)
def test_3d_renumber(dtype):
  bits = np.dtype(dtype).itemsize * 8
  big = (2 ** (bits - 1)) - 1 # cover ints and uints
  data = np.array([
    [
      [big, 0],
      [2, big],
    ],
    [
      [big-5, big-1],
      [big-7, big-3],
    ],
  ], dtype=dtype)

  data2 = np.copy(data, order='C')
  data2, remapdict = fastremap.renumber(data2, preserve_zero=False)

  assert np.all(data2 == [
    [
      [1, 2],
      [3, 1]
    ],
    [ 
      [4, 5],
      [6, 7],
    ],
  ])

  data2 = np.copy(data, order='F')
  data2, remapdict = fastremap.renumber(data2, preserve_zero=False)

  assert np.all(data2 == [
    [
      [1, 5],
      [3, 1]
    ],
    [ 
      [2, 6],
      [4, 7],
    ],
  ])

def test_3d_renumber_dtype_shift():
  big = np.random.randint(0, (2**64)-1, size=(128,128,100), dtype=np.uint64)
  big, remapdict = fastremap.renumber(big, preserve_zero=True, in_place=True)
  assert np.dtype(big.dtype).itemsize <= 4
  assert np.dtype(big.dtype).itemsize > 1

def test_renumber_no_preserve_zero():
  data = np.array([
    [0, 1],
    [1, 2]
  ])
  gt = np.array([
    [0, 1],
    [1, 2]
  ])

  res, remap = fastremap.renumber(data, start=0, preserve_zero=False)
  print(res)
  print(remap)
  assert remap == { 0: 0, 1: 1, 2: 2 }
  assert np.all(res == gt)


@pytest.mark.parametrize("dtype", list(DTYPES) + [ np.float32, np.float64 ])
def test_remap_1d(dtype):
  empty = fastremap.remap([], {})
  assert len(empty) == 0

  data = np.array([1, 2, 2, 2, 3, 4, 5], dtype=dtype)
  remap = {
    1: 10,
    2: 30,
    3: 15,
    4: 0,
    5: 5,
  }

  result = fastremap.remap(np.copy(data), remap, preserve_missing_labels=False)
  assert np.all(result == [10, 30, 30, 30, 15, 0, 5])

  del remap[2]
  try:
    result = fastremap.remap(np.copy(data), remap, preserve_missing_labels=False)
    assert False
  except KeyError:
    pass 

  result = fastremap.remap(np.copy(data), remap, preserve_missing_labels=True)
  assert np.all(result == [10, 2, 2, 2, 15, 0, 5])

@pytest.mark.parametrize("dtype", DTYPES)
def test_remap_2d(dtype):
  data = np.array([[1, 2, 3, 4, 5], [5, 4, 3, 2, 1]], dtype=dtype)
  remap = {
    1: 10,
    2: 30,
    3: 15,
    4: 0,
    5: 5,
  }

  result = fastremap.remap(np.copy(data), remap, preserve_missing_labels=False)
  assert np.all(result == [[10, 30, 15, 0, 5], [5, 0, 15, 30, 10]])

  del remap[2]
  try:
    result = fastremap.remap(np.copy(data), remap, preserve_missing_labels=False)
    assert False
  except KeyError:
    pass 

  result = fastremap.remap(np.copy(data), remap, preserve_missing_labels=True)
  assert np.all(result == [[10, 2, 15, 0, 5], [5, 0, 15, 2, 10]])

def test_remap_broken():
  labels = np.zeros((256, 256, 256), dtype=np.uint32)

  labels[:50, :40, :30] = 2
  labels[50:100, 40:100, 30:80] = 5

  res = fastremap.remap(labels, {5:5}, preserve_missing_labels=True)
  assert np.all(res == labels)

@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("in_place", [ True, False ])
def test_mask(dtype, in_place):
  data = np.arange(100, dtype=dtype)
  data = fastremap.mask(data, [5, 10, 15, 20], in_place=in_place)

  labels, cts = np.unique(data, return_counts=True)
  assert cts[0] == 5 
  assert labels[0] == 0
  assert np.all(cts[1:] == 1)
  assert len(labels == 95)

@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("in_place", [ True, False ])
def test_mask_except(dtype, in_place):
  for value in (0, 7, np.iinfo(dtype).max):
    data = np.arange(100, dtype=dtype)
    data = fastremap.mask_except(
      data, [5, 10, 15, 20], 
      in_place=in_place, value=value
    )

    labels, cts = np.unique(data, return_counts=True)
    print(labels, cts)
    res = { lbl: ct for lbl, ct in zip(labels, cts) }
    assert res == {
      value: 96,
      5: 1,
      10: 1,
      15: 1, 
      20: 1,
    }

@pytest.mark.parametrize("dtype", list(DTYPES) + [ np.float32, np.float64, bool, np.complex64 ])
@pytest.mark.parametrize("dim", [1, 4, 7, 9, 27, 31, 100, 127, 200] )
def test_asfortranarray(dtype, dim):
  x = np.arange(dim**1).reshape((dim)).astype(dtype)
  y = np.copy(x)
  assert np.all(np.asfortranarray(x) == fastremap.asfortranarray(y))

  x = np.arange(dim**2).reshape((dim,dim)).astype(dtype)
  y = np.copy(x)
  assert np.all(np.asfortranarray(x) == fastremap.asfortranarray(y))

  x = np.arange(dim**3).reshape((dim,dim,dim)).astype(dtype)
  y = np.copy(x)
  assert np.all(np.asfortranarray(x) == fastremap.asfortranarray(y))

  x = np.arange(dim**2+dim).reshape((dim,dim+1)).astype(dtype)
  y = np.copy(x)
  assert np.all(np.asfortranarray(x) == fastremap.asfortranarray(y))

  x = np.arange(dim**3+dim*dim).reshape((dim,dim+1,dim)).astype(dtype)
  y = np.copy(x)
  assert np.all(np.asfortranarray(x) == fastremap.asfortranarray(y))

  if dim < 100:
    x = np.arange(dim**4).reshape((dim,dim,dim,dim)).astype(dtype)
    y = np.copy(x)
    assert np.all(np.asfortranarray(x) == fastremap.asfortranarray(y))

    x = np.arange(dim**4 + dim*dim*dim).reshape((dim+1,dim,dim,dim)).astype(dtype)
    y = np.copy(x)
    assert np.all(np.asfortranarray(x) == fastremap.asfortranarray(y))


@pytest.mark.parametrize("dtype", list(DTYPES) + [ np.float32, np.float64, bool, np.complex64 ])
@pytest.mark.parametrize("dim", [1, 4, 7, 9, 27, 31, 100, 127, 200] )
def test_ascontiguousarray(dtype, dim):
  x = np.arange(dim**2).reshape((dim,dim), order='F').astype(dtype)
  y = np.copy(x, order='F')
  assert np.all(np.ascontiguousarray(x) == fastremap.ascontiguousarray(y))

  x = np.arange(dim**3).reshape((dim,dim,dim), order='F').astype(dtype)
  y = np.copy(x, order='F')
  assert np.all(np.ascontiguousarray(x) == fastremap.ascontiguousarray(y))

  x = np.arange(dim**2+dim).reshape((dim,dim+1), order='F').astype(dtype)
  y = np.copy(x, order='F')
  assert np.all(np.ascontiguousarray(x) == fastremap.ascontiguousarray(y))

  x = np.arange(dim**3+dim*dim).reshape((dim,dim+1,dim), order='F').astype(dtype)
  y = np.copy(x, order='F')
  assert np.all(np.ascontiguousarray(x) == fastremap.ascontiguousarray(y))

  if dim < 100:
    x = np.arange(dim**4).reshape((dim,dim,dim,dim)).astype(dtype)
    y = np.copy(x, order='F')
    assert np.all(np.ascontiguousarray(x) == fastremap.ascontiguousarray(y))

    x = np.arange(dim**4 + dim*dim*dim).reshape((dim+1,dim,dim,dim)).astype(dtype)
    y = np.copy(x, order='F')
    assert np.all(np.ascontiguousarray(x) == fastremap.ascontiguousarray(y))

@pytest.mark.parametrize("dtype", [ np.uint8, np.uint16, np.uint32, np.uint64 ])
def test_fit_dtype_uint(dtype):
  assert fastremap.fastremap._fit_dtype(dtype, 0) == np.uint8
  assert fastremap.fastremap._fit_dtype(dtype, 255) == np.uint8
  assert fastremap.fastremap._fit_dtype(dtype, 256) == np.uint16
  assert fastremap.fastremap._fit_dtype(dtype, 10000) == np.uint16
  assert fastremap.fastremap._fit_dtype(dtype, 2**16 - 1) == np.uint16
  assert fastremap.fastremap._fit_dtype(dtype, 2**16) == np.uint32
  assert fastremap.fastremap._fit_dtype(dtype, 2**32) == np.uint64
  assert fastremap.fastremap._fit_dtype(dtype, 2**64 - 1) == np.uint64

  try:
    fastremap.fastremap._fit_dtype(dtype, -1)
    assert False 
  except ValueError:
    pass

  try:
    fastremap.fastremap._fit_dtype(dtype, 2**64)
  except ValueError:
    pass

@pytest.mark.parametrize("dtype", [ np.int8, np.int16, np.int32, np.int64 ])
def test_fit_dtype_int(dtype):
  assert fastremap.fastremap._fit_dtype(dtype, 0) == np.int8
  assert fastremap.fastremap._fit_dtype(dtype, 127) == np.int8
  assert fastremap.fastremap._fit_dtype(dtype, -128) == np.int8
  assert fastremap.fastremap._fit_dtype(dtype, 128) == np.int16
  assert fastremap.fastremap._fit_dtype(dtype, 10000) == np.int16
  assert fastremap.fastremap._fit_dtype(dtype, 2**15 - 1) == np.int16
  assert fastremap.fastremap._fit_dtype(dtype, 2**15) == np.int32
  assert fastremap.fastremap._fit_dtype(dtype, 2**32) == np.int64
  assert fastremap.fastremap._fit_dtype(dtype, 2**63 - 1) == np.int64

  try:
    fastremap.fastremap._fit_dtype(dtype, 2**63)
  except ValueError:
    pass

  try:
    fastremap.fastremap._fit_dtype(dtype, -2**63)
  except ValueError:
    pass

@pytest.mark.parametrize("dtype", [ np.float16, np.float32, np.float64 ])
def test_fit_dtype_float(dtype):
  assert fastremap.fastremap._fit_dtype(dtype, 0) == np.float32
  assert fastremap.fastremap._fit_dtype(dtype, 127) == np.float32
  assert fastremap.fastremap._fit_dtype(dtype, 128) == np.float32
  assert fastremap.fastremap._fit_dtype(dtype, 10000) == np.float32
  assert fastremap.fastremap._fit_dtype(dtype, 2**15 - 1) == np.float32
  assert fastremap.fastremap._fit_dtype(dtype, 2**15) == np.float32
  assert fastremap.fastremap._fit_dtype(dtype, 2**32) == np.float32
  assert fastremap.fastremap._fit_dtype(dtype, 2**63 - 1) == np.float32
  assert fastremap.fastremap._fit_dtype(dtype, -2**63) == np.float32
  assert fastremap.fastremap._fit_dtype(dtype, 2**128) == np.float64

  assert fastremap.fastremap._fit_dtype(dtype, 0, exotics=True) == np.float16
  assert fastremap.fastremap._fit_dtype(dtype, 127, exotics=True) == np.float16
  assert fastremap.fastremap._fit_dtype(dtype, 128, exotics=True) == np.float16
  assert fastremap.fastremap._fit_dtype(dtype, 10000, exotics=True) == np.float16
  assert fastremap.fastremap._fit_dtype(dtype, 2**15 - 1, exotics=True) == np.float16
  assert fastremap.fastremap._fit_dtype(dtype, 2**15, exotics=True) == np.float16
  assert fastremap.fastremap._fit_dtype(dtype, 2**32, exotics=True) == np.float32
  assert fastremap.fastremap._fit_dtype(dtype, 2**63 - 1, exotics=True) == np.float32
  assert fastremap.fastremap._fit_dtype(dtype, -2**63, exotics=True) == np.float32

@pytest.mark.parametrize("dtype", [ np.csingle, np.cdouble ])
@pytest.mark.parametrize("sign", [ 1, -1, 1j, -1j ])
def test_fit_dtype_float(dtype, sign):
  assert fastremap.fastremap._fit_dtype(dtype, sign * 0+0j) == np.csingle
  assert fastremap.fastremap._fit_dtype(dtype, sign * 127) == np.csingle
  assert fastremap.fastremap._fit_dtype(dtype, sign * 127) == np.csingle
  assert fastremap.fastremap._fit_dtype(dtype, sign * 128) == np.csingle
  assert fastremap.fastremap._fit_dtype(dtype, sign * 128) == np.csingle
  assert fastremap.fastremap._fit_dtype(dtype, sign * 10000) == np.csingle
  assert fastremap.fastremap._fit_dtype(dtype, sign * 10000) == np.csingle
  assert fastremap.fastremap._fit_dtype(dtype, sign * 2**15 - 1) == np.csingle
  assert fastremap.fastremap._fit_dtype(dtype, sign * 2**15) == np.csingle
  assert fastremap.fastremap._fit_dtype(dtype, sign * 2**32) == np.csingle
  assert fastremap.fastremap._fit_dtype(dtype, sign * 2**63 - 1) == np.csingle
  assert fastremap.fastremap._fit_dtype(dtype, -2**63) == np.csingle
  
  try:
    fastremap.fastremap._fit_dtype(dtype, sign * 2**128)
    assert False
  except ValueError:
    pass

  assert fastremap.fastremap._fit_dtype(dtype, sign * 2**128, exotics=True) == np.cdouble

def test_minmax():
  volume = np.random.randint(-500, 500, size=(128,128,128))
  minval, maxval = fastremap.minmax(volume)
  assert minval == np.min(volume)
  assert maxval == np.max(volume)

@pytest.mark.parametrize("dtype", DTYPES)
def test_unique_axis_0(dtype):
  arr = np.array([
    [0,1],
    [0,2],
    [0,3],
    [0,2],
    [1,2],
    [2,3],
  ], dtype=dtype)

  res = fastremap.unique(arr, axis=0)
  ans = np.unique(arr, axis=0)

  assert np.all(res == ans)

def test_unique_axis_0_random():
  arr = np.random.randint(0,100000, size=[1000000,2], dtype=np.uint32)
  a1 = fastremap.unique(arr, axis=0)
  a2 = np.unique(arr, axis=0)
  assert np.all(a1 == a2)


@pytest.mark.parametrize("order", [ "C", "F" ])
def test_unique(order):
  def reorder(arr):
    if order == "F":
      return np.asfortranarray(arr)
    return np.ascontiguousarray(arr)

  assert len(fastremap.unique(np.array([], dtype=np.uint8))) == 0

  # array_unique
  labels = reorder(np.random.randint(0, 500, size=(128,128,128)))
  uniq_np, idx_np, inv_np, cts_np = np.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  uniq_fr, idx_fr, inv_fr, cts_fr = fastremap.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  assert np.all(uniq_np == uniq_fr)
  assert np.all(inv_np == inv_fr)
  assert np.all(cts_np == cts_fr)
  assert np.all(labels.flatten()[idx_np] == labels.flatten()[idx_fr])

  labels = reorder(np.random.randint(0, 500, size=(128,128,128)))
  uniq_np, idx_np, inv_np, cts_np = np.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  uniq_fr, idx_fr, cts_fr, inv_fr = fastremap.fastremap._unique_via_array(labels.flatten(), np.max(labels), return_index=True, return_inverse=True)
  assert np.all(uniq_np == uniq_fr)
  assert np.all(inv_np.flatten() == inv_fr)
  assert np.all(cts_np == cts_fr)
  assert np.all(labels.flatten()[idx_np] == labels.flatten()[idx_fr])

  # array_unique + shift
  labels = reorder(np.random.randint(-500, 500, size=(128,128,128)))
  uniq_np, idx_np, inv_np, cts_np = np.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  uniq_fr, idx_fr, inv_fr, cts_fr = fastremap.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  assert np.all(uniq_np == uniq_fr)
  assert np.all(inv_np == inv_fr)
  assert np.all(cts_np == cts_fr)
  assert np.all(labels.flatten()[idx_np] == labels.flatten()[idx_fr])

  labels = reorder(np.random.randint(-500, 500, size=(128,128,128)))
  uniq_np, idx_np, inv_np, cts_np = np.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  uniq_fr, idx_fr, cts_fr, inv_fr = fastremap.fastremap._unique_via_shifted_array(labels.flatten(), return_index=True, return_inverse=True)
  assert np.all(uniq_np == uniq_fr)
  assert np.all(inv_np.flatten() == inv_fr)
  assert np.all(cts_np == cts_fr)
  assert np.all(labels.flatten()[idx_np] == labels.flatten()[idx_fr])

  # array_unique + shift
  labels = reorder(np.random.randint(128**3 - 500, 128**3 + 500, size=(128,128,128)))
  uniq_np, idx_np, inv_np, cts_np = np.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  uniq_fr, idx_fr, inv_fr, cts_fr = fastremap.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  assert np.all(uniq_np == uniq_fr)
  assert np.all(inv_np == inv_fr)
  assert np.all(cts_np == cts_fr)
  assert np.all(labels.flatten()[idx_np] == labels.flatten()[idx_fr])

  # array_unique + shift
  labels = reorder(np.random.randint(128**3 - 500, 128**3 + 500, size=(128,128,128)))
  uniq_np, idx_np, inv_np, cts_np = np.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  uniq_fr, idx_fr, cts_fr, inv_fr = fastremap.fastremap._unique_via_shifted_array(labels.flatten(), return_index=True, return_inverse=True)
  assert np.all(uniq_np == uniq_fr)
  assert np.all(inv_np.flatten() == inv_fr)
  assert np.all(cts_np == cts_fr)
  assert np.all(labels.flatten()[idx_np] == labels.flatten()[idx_fr])

  # renumber + array_unique
  labels = reorder(np.random.randint(0, 1, size=(128,128,128)))
  labels[0,0,0] = 128**3 + 10
  uniq_np, idx_np, inv_np, cts_np = np.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  uniq_fr, idx_fr, inv_fr, cts_fr = fastremap.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  assert np.all(uniq_np == uniq_fr)
  assert np.all(inv_np == inv_fr)
  assert np.all(cts_np == cts_fr)  
  assert np.all(labels.flatten()[idx_np] == labels.flatten()[idx_fr])

  labels = reorder(np.random.randint(0, 1, size=(128,128,128)))
  labels[0,0,0] = 128**3 + 10
  uniq_np, idx_np, inv_np, cts_np = np.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  uniq_fr, idx_fr, cts_fr, inv_fr = fastremap.fastremap._unique_via_renumber(labels.flatten(), return_index=True, return_inverse=True)
  assert np.all(uniq_np == uniq_fr)
  assert np.all(inv_np.flatten() == inv_fr)
  assert np.all(cts_np == cts_fr)  
  assert np.all(labels.flatten()[idx_np] == labels.flatten()[idx_fr])

  # sort
  labels = reorder(np.random.randint(-1000, 128**3, size=(100,100,100)))
  uniq_np, idx_np, inv_np, cts_np = np.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  uniq_fr, idx_fr, inv_fr, cts_fr = fastremap.unique(labels, return_counts=True, return_index=True, return_inverse=True)
  assert np.all(uniq_np == uniq_fr)
  assert np.all(inv_np == inv_fr)
  assert np.all(cts_np == cts_fr)  
  assert np.all(labels.flatten()[idx_np] == labels.flatten()[idx_fr])

  labels = reorder(np.random.randint(-1000, 128**3, size=(100,100,100)))
  uniq_np, cts_np = np.unique(labels, return_counts=True)
  uniq_fr, cts_fr = fastremap.fastremap._unique_via_sort(labels.flatten())
  assert np.all(uniq_np == uniq_fr)
  assert np.all(cts_np == cts_fr)

  labels = [ 1, 1, 2, 3, 2, 4, 3 ]
  uniq = fastremap.unique(labels)
  assert np.all(uniq == np.array([1,2,3,4]))


def test_renumber_remap():
  labels = np.random.randint(-500, 500, size=(128,128,128)).astype(np.int64)
  new_labels, remap = fastremap.renumber(labels, in_place=False)
  remap = { v:k for k,v in remap.items() }
  new_labels = fastremap.remap(new_labels, remap, in_place=True)
  assert np.all(labels == new_labels)
  assert new_labels.dtype in (np.int8, np.int16)
  assert labels.dtype == np.int64

@pytest.mark.parametrize("dtype_cc", DTYPES)
@pytest.mark.parametrize("dtype_p", DTYPES)
def test_component_map(dtype_cc, dtype_p):
  shape = (128,128,128)
  cc_labels = np.random.randint(0, 100, size=shape).astype(dtype_cc)
  parent_labels = (cc_labels + 1).astype(dtype_p)

  mapping = fastremap.component_map(cc_labels, parent_labels)
  for k,v in mapping.items():
    assert k == v - 1

  mapping = fastremap.component_map([ 1, 2, 3, 4 ], [ 5, 5, 6, 7 ])
  assert mapping == { 1: 5, 2: 5, 3: 6, 4: 7 }

  mapping = fastremap.component_map([], [])


@pytest.mark.parametrize("dtype_cc", DTYPES)
@pytest.mark.parametrize("dtype_p", DTYPES)
def test_inverse_component_map(dtype_cc, dtype_p):
  mapping = fastremap.inverse_component_map([ 1, 2, 1, 3 ], [ 4, 4, 5, 6 ])
  assert mapping == { 1: [ 4, 5 ], 2: [ 4 ], 3: [ 6 ] }

  mapping = fastremap.inverse_component_map([ 1, 1, 1, 3 ], [ 4, 4, 5, 6 ]) 
  assert mapping == {1: [4, 5], 3: [6]}

  mapping = fastremap.inverse_component_map([ 1, 1, 1, 1 ], [ 4, 2, 5, 6 ]) 
  assert mapping == {1: [2, 4, 5, 6]}

  mapping = fastremap.inverse_component_map([], [])

def test_point_cloud():
  x = np.ones((2,2,2), dtype=np.uint8)
  ptc = fastremap.point_cloud(x)
  assert len(ptc) == 1

  gt = np.array([
    [0, 0, 0],
    [0, 0, 1],
    [0, 1, 0],
    [0, 1, 1],
    [1, 0, 0],
    [1, 0, 1],
    [1, 1, 0],
    [1, 1, 1],
  ])
  assert np.all(ptc[1] == gt)

  x[0,0,0] = 2
  ptc = fastremap.point_cloud(x)
  assert len(ptc) == 2
  assert np.all(ptc[1] == gt[1:,:])
  assert np.all(ptc[2] == gt[:1,:])

  x[1,1,1] = 3
  ptc = fastremap.point_cloud(x)
  assert len(ptc) == 3
  assert np.all(ptc[1] == gt[1:7,:])
  assert np.all(ptc[2] == gt[:1,:])
  assert np.all(ptc[3] == gt[7:,:])  

  x = np.ones((0,0,0), dtype=np.uint8)
  ptc = fastremap.point_cloud(x)
  assert len(ptc) == 0


@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("input_order", ['C','F'])
@pytest.mark.parametrize("output_order", ['F', 'C'])
@pytest.mark.parametrize("size", [16,64,128])
@pytest.mark.parametrize("chunk_size", [4,8,16])
def test_tobytes(size, chunk_size, dtype, input_order, output_order):
  cs = chunk_size
  image = np.arange(size*size*size, dtype=dtype).reshape((size,size,size), order=input_order)
  
  res1 = fastremap.tobytes(image, (cs,cs,cs), order=output_order)

  N = size // cs

  res2 = []
  for z in range(N):
    for y in range(N):
      for x in range(N):
        cutout = image[x*cs:(x+1)*cs, y*cs:(y+1)*cs, z*cs:(z+1)*cs]
        res2.append(cutout.tobytes(output_order))

  for i, (enc1, enc2) in enumerate(zip(res1, res2)):
    assert enc1 == enc2, i

@pytest.mark.parametrize("order", ['C','F'])
def test_tobytes_misaligned(order):
  size = 128
  cs = 32
  image = np.arange(size*size*size, dtype=np.uint8).reshape((size,size,size), order=order)

  res1 = fastremap.tobytes(image, (cs,cs,cs))

  cs = 17
  try:
    res1 = fastremap.tobytes(image, (cs,cs,cs))
    assert False
  except ValueError:
    pass

def test_narrow_dtype():
  assert fastremap.narrow_dtype(np.uint64) == np.uint32
  assert fastremap.narrow_dtype(np.uint32) == np.uint16
  assert fastremap.narrow_dtype(np.uint16) == np.uint8
  assert fastremap.narrow_dtype(np.uint8) == np.uint8

  assert fastremap.narrow_dtype(np.int64) == np.int32
  assert fastremap.narrow_dtype(np.int32) == np.int16
  assert fastremap.narrow_dtype(np.int16) == np.int8
  assert fastremap.narrow_dtype(np.int8) == np.int8

  assert fastremap.narrow_dtype(np.uint64, exotics=True) == np.uint32
  assert fastremap.narrow_dtype(np.uint32, exotics=True) == np.uint16
  assert fastremap.narrow_dtype(np.uint16, exotics=True) == np.uint8
  assert fastremap.narrow_dtype(np.uint8, exotics=True) == np.uint8

  assert fastremap.narrow_dtype(np.int64, exotics=True) == np.int32
  assert fastremap.narrow_dtype(np.int32, exotics=True) == np.int16
  assert fastremap.narrow_dtype(np.int16, exotics=True) == np.int8
  assert fastremap.narrow_dtype(np.int8, exotics=True) == np.int8

  assert fastremap.narrow_dtype(np.float64) == np.float32
  assert fastremap.narrow_dtype(np.float32) == np.float32

  assert fastremap.narrow_dtype(np.float64, exotics=True) == np.float32
  assert fastremap.narrow_dtype(np.float32, exotics=True) == np.float16

  assert fastremap.narrow_dtype(np.complex128) == np.complex64
  assert fastremap.narrow_dtype(np.complex64) == np.complex64

def test_widen_dtype():
  assert fastremap.widen_dtype(np.uint64) == np.uint64
  assert fastremap.widen_dtype(np.uint32) == np.uint64
  assert fastremap.widen_dtype(np.uint16) == np.uint32
  assert fastremap.widen_dtype(np.uint8) == np.uint16

  assert fastremap.widen_dtype(np.int64) == np.int64
  assert fastremap.widen_dtype(np.int32) == np.int64
  assert fastremap.widen_dtype(np.int16) == np.int32
  assert fastremap.widen_dtype(np.int8) == np.int16

  assert fastremap.widen_dtype(np.float64, exotics=True) == np.longdouble
  assert fastremap.widen_dtype(np.float64) == np.float64
  assert fastremap.widen_dtype(np.float32) == np.float64
  assert fastremap.widen_dtype(np.float16) == np.float32

  assert fastremap.widen_dtype(np.complex64) == np.complex64
  assert fastremap.widen_dtype(np.complex64, exotics=True) == np.complex128
  assert fastremap.widen_dtype(np.complex128, exotics=True) == np.clongdouble
  assert fastremap.widen_dtype(np.clongdouble, exotics=True) == np.clongdouble

def test_unique_order():
  data = np.array([
    1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704,
    1704, 1704, 1704, 1704, 1704, 1704,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704,
    1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704,
    1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704,
    1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704,
    1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704,
    1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704,
    1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704,
    1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704,
    1704, 1704, 1704, 1704, 1704,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704,
     1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704,
     1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704,
     1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704, 1704,
     1704, 1704, 1704,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106, 1873,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106, 1873, 1873,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
      106, 1873, 1873,  106,  106,  106,  106,  106,  106,  106,  106,
      106,  106,  106,  106,  106,  106,  106
    ])
  tup = fastremap.unique(data, return_counts=True, return_index=True, return_inverse=True)
  tup2 = np.unique(data, return_counts=True, return_index=True, return_inverse=True)

  for res, ans in zip(tup, tup2):
    assert np.all(res == ans)



from .fastremap import (
    ascontiguousarray,
    asfortranarray,
    component_map,
    foreground,
    indices,
    inverse_component_map,
    mask,
    mask_except,
    minmax,
    narrow_dtype,
    pixel_pairs,
    point_cloud,
    refit,
    remap,
    remap_from_array,
    remap_from_array_kv,
    renumber,
    tobytes,
    transpose,
    unique,
    widen_dtype,
)

__all__ = [
    "ascontiguousarray",
    "asfortranarray",
    "component_map",
    "foreground",
    "indices",
    "inverse_component_map",
    "mask",
    "mask_except",
    "minmax",
    "narrow_dtype",
    "pixel_pairs",
    "point_cloud",
    "refit",
    "remap",
    "remap_from_array",
    "remap_from_array_kv",
    "renumber",
    "tobytes",
    "transpose",
    "unique",
    "widen_dtype",
]



import fastremap
import numpy as np 

x = np.ones((512,512,512), dtype=np.float32)
x = fastremap.asfortranarray(x)
print(x)
print(x.flags)
print(x.strides)

print(x.dtype)



# @profile
# def run():
#   x = np.ones( (512,512,512), dtype=np.uint32, order='C')
#   x += 1
#   print(x.strides, x.flags)
#   y = np.asfortranarray(x)
#   print(x.strides, x.flags)

#   print("done.")

# run()